!MNH_LIC Copyright 2009-2023 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
!     ##########################
      MODULE MODI_MEAN_FIELD
!     ##########################
!
!
INTERFACE

      SUBROUTINE MEAN_FIELD( PUT, PVT, PWT, PTHT, PTKET, PPABST, PRT, PSVT )

REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PUT, PVT, PWT   ! variables
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PTHT, PTKET   ! variables
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PPABST   ! variables
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRT      ! variables
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PSVT   ! Passive scalar variables

END SUBROUTINE MEAN_FIELD

END INTERFACE

END MODULE MODI_MEAN_FIELD
!
!     ######################################################################
      SUBROUTINE MEAN_FIELD( PUT, PVT, PWT, PTHT, PTKET, PPABST, PRT, PSVT )
!     ######################################################################
!
!!****  *MEAN_FIELD * -
!!
!!    PURPOSE
!!    -------
!
!!**  METHOD
!!    ------
!!
!!    REFERENCE
!!    ---------
!!      
!!
!!    AUTHOR
!!    ------
!!     P. Aumond 
!!
!!    MODIFICATIONS
!!    -------------
!!      Original       07/2009
!!      (C.Lac)        09/2016 Max values
!!      (PA.Joulin)    12/2020 Wind turbine variables
!!      (R. Schoetter) 12/2021 adds humidity and other mean diagnostics
!!      (E. Jezequel)  11/2022 Welford algorithm and covariances
!!      (H. Toumi)     09/2022: add ADR
!!      (M. Mandement) 01/2024: add max 10 m wind gust speed formulations
!!---------------------------------------------------------------
!
!
!*       0.    DECLARATIONS
!              ------------
!
USE MODE_ll
USE MODD_CONF_n, ONLY: LUSERV
USE MODD_MEAN_FIELD_n
USE MODD_PARAM_n
USE MODD_MEAN_FIELD
USE MODD_CST
USE MODD_PASPOL
! To play with MPI
USE MODD_ARGSLIST_ll, ONLY: LIST_ll
!
! * EOL
USE MODD_EOL_MAIN, ONLY: LMAIN_EOL, CMETH_EOL, NMODEL_EOL
USE MODD_EOL_SHARED_IO, ONLY: XTHRUT, XTORQT, XPOWT, XAOA_GLB
USE MODD_EOL_SHARED_IO, ONLY: XTHRU_SUM, XTORQ_SUM, XPOW_SUM, XAOA_SUM
USE MODD_EOL_ADR, ONLY: XFAERO_RA_GLB, XFAERO_RA_SUM
USE MODD_EOL_ADR, ONLY: XFAERO_BLEQ_RA_GLB, XFAERO_BLEQ_RA_SUM
USE MODD_EOL_ADR, ONLY: XAOA_BLEQ_GLB, XAOA_BLEQ_SUM
USE MODD_EOL_ALM, ONLY: XFAERO_RE_SUM, XFAERO_RE_GLB
!
USE MODI_SHUMAN,  ONLY: MXF, MYF, MZF
!
USE MODE_MODELN_HANDLER
USE MODE_THERMO
USE MODI_UPDATE_WELFORD
! 
USE MODI_SHUMAN
USE MODI_GRADIENT_M
USE MODI_GRADIENT_U
USE MODI_GRADIENT_V
USE MODI_GRADIENT_W
USE MODD_METRICS_n, ONLY: XDXX, XDYY, XDZX, XDZY, XDZZ
USE MODD_FIELD_n, ONLY: XRT
USE MODD_CST, ONLY: XRD, XRV, XG
USE MODD_GRID_n, ONLY: XZS, XZZ, XZHAT, XZHATM
USE MODD_CONF_n, ONLY: NRR, NRRL, NRRI
!

IMPLICIT NONE

!*       0.1   Declarations of dummy arguments :
!
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PUT, PVT, PWT   ! variables
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PTHT, PTKET   ! variables
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PPABST   ! variables
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRT      ! variables
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PSVT

!
!*       0.2   Declarations of local variables :
INTEGER           :: IIU,IJU,IKU,IIB,IJB,IKB,IIE,IJE,IKE ! Arrays bounds
REAL, DIMENSION(:,:,:), ALLOCATABLE :: ZUMEAN_OLD
REAL, DIMENSION(:,:,:), ALLOCATABLE :: ZWMEAN_OLD
!
REAL, DIMENSION(:,:,:), ALLOCATABLE :: ZRT
REAL, DIMENSION(:,:,:), ALLOCATABLE :: ZQSAT_W
REAL, DIMENSION(:,:,:), ALLOCATABLE :: ZQSAT_I
REAL, DIMENSION(:,:,:), ALLOCATABLE :: ZQACT
REAL, DIMENSION(:,:,:), ALLOCATABLE :: ZRH_W
REAL, DIMENSION(:,:,:), ALLOCATABLE :: ZRH_I
REAL, DIMENSION(:,:,:), ALLOCATABLE :: ZRH_P
REAL, DIMENSION(:,:,:), ALLOCATABLE :: ZFRAC
REAL, DIMENSION(SIZE(PUT,1),SIZE(PUT,2),SIZE(PUT,3)) ::  ZAUX_WIFF
REAL, DIMENSION(SIZE(PUT,1),SIZE(PUT,2),SIZE(PUT,3)) ::  ZAUX_WIDD
REAL, DIMENSION(:,:),   ALLOCATABLE :: ZRH_W_MAXCOL
REAL, DIMENSION(:,:),   ALLOCATABLE :: ZRH_I_MAXCOL
REAL, DIMENSION(:,:),   ALLOCATABLE :: ZRH_P_MAXCOL
INTEGER           :: JI,JJ,JK, JLOOP   ! Loop indexes
REAL, DIMENSION(SIZE(PUT,1),SIZE(PUT,2),SIZE(PUT,3)) ::  ZTEMPT
REAL, DIMENSION(SIZE(PUT,1),SIZE(PUT,2),SIZE(PUT,3)) :: ZTHV ! 
REAL, DIMENSION(SIZE(PUT,1),SIZE(PUT,2),SIZE(PUT,3)) ::  ZVOX, ZVOY, ZVOZ ! 
REAL, DIMENSION(SIZE(PUT,1),SIZE(PUT,2),SIZE(PUT,3)) ::  ZWM ! 
REAL, DIMENSION(SIZE(PUT,1),SIZE(PUT,2)) :: ZWORK1,ZWORK2,ZWORK3,ZWORK4,ZWORK5,ZWORK6,ZWORK7
REAL, DIMENSION(SIZE(PUT,1),SIZE(PUT,2)) :: ZCURRENT_ZON10M,ZCURRENT_MER10M
REAL, DIMENSION(SIZE(PUT,1),SIZE(PUT,2)) :: ZCURRENT_TKE10M,ZCURRENT_TKE20M
REAL, DIMENSION(SIZE(PUT,1),SIZE(PUT,2)) ::  ZMSLP ! 
REAL, DIMENSION(SIZE(PUT,1),SIZE(PUT,2)) ::  ZUH ! 
REAL    :: ZRV_OV_RD, ZGAMREF ! 
!
INTEGER :: IMI !Current model index
!
! For communication
TYPE(LIST_ll), POINTER :: TZFIELDS_Cov_ll  ! Field list of Wind for exchange beofre covariances
INTEGER                :: IINFO            ! Info integer
!
!-----------------------------------------------------------------------
!
!*       0.     ARRAYS BOUNDS INITIALIZATION
!
IIU=SIZE(PTHT,1)
IJU=SIZE(PTHT,2)
IKU=SIZE(PTHT,3)
CALL GET_INDICE_ll (IIB,IJB,IIE,IJE)
IKB=1+JPVEXT
IKE=IKU-JPVEXT
!
!-----------------------------------------------------------------------
!
!*       1. MEAN
!
ZTEMPT = PTHT*(((PPABST)/XP00)**(XRD/XCPD))
ZCURRENT_ZON10M = 0.
ZCURRENT_MER10M = 0.
ZCURRENT_TKE10M = 0.
ZCURRENT_TKE20M = 0.
!
!
IF(LUSERV) THEN
  ALLOCATE( ZRT    (IIU, IJU, IKU) )
  ALLOCATE( ZQSAT_W(IIU, IJU, IKU) )
  ALLOCATE( ZQSAT_I(IIU, IJU, IKU) )
  ALLOCATE( ZQACT  (IIU, IJU, IKU) )
  ALLOCATE( ZRH_W  (IIU, IJU, IKU) )
  ALLOCATE( ZRH_I  (IIU, IJU, IKU) )
  ALLOCATE( ZRH_P  (IIU, IJU, IKU) )
  ALLOCATE( ZFRAC  (IIU, IJU, IKU) )
  ALLOCATE( ZRH_W_MAXCOL( IIU, IJU) )
  ALLOCATE( ZRH_I_MAXCOL( IIU, IJU) )
  ALLOCATE( ZRH_P_MAXCOL( IIU, IJU) )

  ! Calculation of saturation specific humidity over water/ice
  !
  ZQSAT_W = QSAT (ZTEMPT, PPABST)
  ZQSAT_I = QSATI(ZTEMPT, PPABST)
  !
  ! Conversion mixing ratio -> specfic humidity
  !
  ZRT(:,:,:) = -9999.0
  WHERE (PRT(:,:,:).LT.1.0E-6)
     ZRT(:,:,:) = 1.0E-6
  ELSEWHERE
     ZRT(:,:,:) = PRT(:,:,:)
  ENDWHERE
  !
  ZQACT(:,:,:) = 1.0 / ( 1.0 + 1.0/ZRT(:,:,:) )
  !
  ! Calculation of relative humidity with respect to water/ice
  !
  ZRH_W(:,:,:) = 100.0*ZQACT(:,:,:)/ZQSAT_W(:,:,:)
  ZRH_I(:,:,:) = 100.0*ZQACT(:,:,:)/ZQSAT_I(:,:,:)
  !
  ! Fractional partitioning between liquid and solid cloud water
  ! as assumed in condensations
  !
  ZFRAC(:,:,:) = ( XTT - ZTEMPT(:,:,:) ) / 20.
  ZFRAC(:,:,:) = MAX( 0., MIN(1., ZFRAC(:,:,:) ) )
  !
  ! Calculation of weighted average between water and ice value
  !
  ZRH_P(:,:,:) = ZFRAC(:,:,:) * ZRH_I(:,:,:) + (1.0-ZFRAC(:,:,:)) * ZRH_W(:,:,:)
  !
  ! Calculation of the column maximum of relative humidity
  !
  ZRH_W_MAXCOL(:,:) = MAXVAL(ZRH_W(:,:,:),DIM=3)
  ZRH_I_MAXCOL(:,:) = MAXVAL(ZRH_I(:,:,:),DIM=3)
  ZRH_P_MAXCOL(:,:) = MAXVAL(ZRH_P(:,:,:),DIM=3)
  XQ_MEAN     = XQ_MEAN + ZQACT
  XRH_W_MEAN    = XRH_W_MEAN + ZRH_W
  XRH_I_MEAN    = XRH_I_MEAN + ZRH_I
  XRH_P_MEAN    = XRH_P_MEAN + ZRH_P
  XRH_W_MAXCOL_MEAN = XRH_W_MAXCOL_MEAN + ZRH_W_MAXCOL
  XRH_I_MAXCOL_MEAN = XRH_I_MAXCOL_MEAN + ZRH_I_MAXCOL
  XRH_P_MAXCOL_MEAN = XRH_P_MAXCOL_MEAN + ZRH_P_MAXCOL

  DEALLOCATE( ZRT, ZQSAT_W, ZQSAT_I, ZQACT, ZRH_W, ZRH_I, ZRH_P, ZFRAC )
  DEALLOCATE( ZRH_W_MAXCOL, ZRH_I_MAXCOL, ZRH_P_MAXCOL )
END IF

   IF (LPASPOL)  XSVT_MEAN  = PSVT + XSVT_MEAN
   IF (CTURB/='NONE') XTKEM_MEAN = PTKET + XTKEM_MEAN
!
!  Wind turbine variables
   IMI = GET_CURRENT_MODEL_INDEX()
   IF (LMAIN_EOL .AND. IMI==NMODEL_EOL) THEN
    SELECT CASE(CMETH_EOL)
     CASE('ADNR') ! Actuator Disc Non-Rotating
      XTHRU_SUM       = XTHRUT        + XTHRU_SUM
     CASE('ADR') ! Actuator Disc with Rotation
      XAOA_SUM        = XAOA_GLB      + XAOA_SUM
      XFAERO_RA_SUM   = XFAERO_RA_GLB + XFAERO_RA_SUM
      XTHRU_SUM       = XTHRUT        + XTHRU_SUM
      XTORQ_SUM       = XTORQT        + XTORQ_SUM
      XPOW_SUM        = XPOWT         + XPOW_SUM
      XAOA_BLEQ_SUM       = XAOA_BLEQ_GLB      + XAOA_BLEQ_SUM
      XFAERO_BLEQ_RA_SUM  = XFAERO_BLEQ_RA_GLB + XFAERO_BLEQ_RA_SUM
     CASE('ALM') ! Actuator Line Method
      XAOA_SUM        = XAOA_GLB      + XAOA_SUM
      XFAERO_RE_SUM   = XFAERO_RE_GLB + XFAERO_RE_SUM
      XTHRU_SUM       = XTHRUT        + XTHRU_SUM
      XTORQ_SUM       = XTORQT        + XTORQ_SUM
      XPOW_SUM        = XPOWT         + XPOW_SUM
    END SELECT
   END IF
!
   MEAN_COUNT = MEAN_COUNT + 1
!
!  Save old mean values for covariance computations
!
   IF (LCOV_FIELD) THEN
      ALLOCATE( ZUMEAN_OLD (IIU, IJU, IKU) )
      ALLOCATE( ZWMEAN_OLD (IIU, IJU, IKU) )
      ZUMEAN_OLD (:,:,:) = XUM_MEAN (:,:,:)
      ZWMEAN_OLD (:,:,:) = XWM_MEAN (:,:,:)
   END IF
!  Welford method for variables whom we compute variances
!
   CALL UPDATE_WELFORD(MEAN_COUNT,XUM_MEAN,XU2_M2,PUT)
   CALL UPDATE_WELFORD(MEAN_COUNT,XVM_MEAN,XV2_M2,PVT)
   CALL UPDATE_WELFORD(MEAN_COUNT,XWM_MEAN,XW2_M2,PWT)
   CALL UPDATE_WELFORD(MEAN_COUNT,XTHM_MEAN,XTH2_M2,PTHT)
   CALL UPDATE_WELFORD(MEAN_COUNT,XTEMPM_MEAN,XTEMP2_M2,ZTEMPT)
   CALL UPDATE_WELFORD(MEAN_COUNT,XPABSM_MEAN,XPABS2_M2,PPABST)
!
!  Welford method for covariances
!
   IF (LCOV_FIELD) THEN 
! 
    NULLIFY(TZFIELDS_Cov_ll)
    CALL ADD3DFIELD_ll( TZFIELDS_Cov_ll,PUT,        'MEAN_FIELD::PUT')
    CALL ADD3DFIELD_ll( TZFIELDS_Cov_ll,PWT,        'MEAN_FIELD::PWT')
    CALL ADD3DFIELD_ll( TZFIELDS_Cov_ll,PVT,        'MEAN_FIELD::PVT')
    CALL ADD3DFIELD_ll( TZFIELDS_Cov_ll,ZUMEAN_OLD, 'MEAN_FIELD::ZUMEAN_OLD')
    CALL ADD3DFIELD_ll( TZFIELDS_Cov_ll,ZWMEAN_OLD, 'MEAN_FIELD::ZWMEAN_OLD')
    CALL ADD3DFIELD_ll( TZFIELDS_Cov_ll,XVM_MEAN,   'MEAN_FIELD::XVM_MEAN')
    CALL ADD3DFIELD_ll( TZFIELDS_Cov_ll,XWM_MEAN,   'MEAN_FIELD::XWM_MEAN')
    CALL UPDATE_HALO_ll(TZFIELDS_Cov_ll,IINFO)
    CALL CLEANLIST_ll(  TZFIELDS_Cov_ll)
!
    XUV_MEAN  = XUV_MEAN  + (MXF(PUT)-MXF(ZUMEAN_OLD))*(MYF(PVT)-MYF(XVM_MEAN))
    XUW_MEAN  = XUW_MEAN  + (MXF(PUT)-MXF(ZUMEAN_OLD))*(MZF(PWT)-MZF(XWM_MEAN))
    XVW_MEAN  = XVW_MEAN  + (MZF(PWT)-MZF(ZWMEAN_OLD))*(MYF(PVT)-MYF(XVM_MEAN))
    XWTH_MEAN = XWTH_MEAN + (MZF(PWT)-MZF(ZWMEAN_OLD))*(PTHT-XTHM_MEAN)
    DEALLOCATE( ZUMEAN_OLD, ZWMEAN_OLD )
   END IF
!
!-----------------------------------------------------------------------
!
!*       2. MAX & MIN
!
  !
  ! Calculation of horizontal wind speed for maximum wind speed diagnostics
  !
  ZAUX_WIFF(:,:,:) = SQRT(MXF(PUT(:,:,:))**2 + MYF(PVT(:,:,:))**2)
  ZAUX_WIDD(:,:,:) = 180.0 + (90.0 - 180.0*ATAN2(MYF(PVT(:,:,:)),MXF(PUT(:,:,:)))/XPI)
  !
  WHERE (ZAUX_WIDD(:,:,:).GT.360.0) ZAUX_WIDD(:,:,:) = ZAUX_WIDD(:,:,:) - 360.0
  !
  DO JK=IKB,IKE
    DO JJ=IJB,IJE
      DO JI=IIB,IIE
        !
        XUM_MAX(JI,JJ,JK) = MAX(XUM_MAX(JI,JJ,JK),PUT(JI,JJ,JK))
        XVM_MAX(JI,JJ,JK) = MAX(XVM_MAX(JI,JJ,JK),PVT(JI,JJ,JK))
        XWM_MAX(JI,JJ,JK) = MAX(XWM_MAX(JI,JJ,JK),PWT(JI,JJ,JK))
        XWM_MIN(JI,JJ,JK) = MIN(XWM_MIN(JI,JJ,JK),PWT(JI,JJ,JK))
        XTHM_MAX(JI,JJ,JK) = MAX(XTHM_MAX(JI,JJ,JK),PTHT(JI,JJ,JK))
        XTEMPM_MAX(JI,JJ,JK) = MAX(XTEMPM_MAX(JI,JJ,JK),ZTEMPT(JI,JJ,JK))
        IF (CTURB/='NONE') XTKEM_MAX(JI,JJ,JK) =  &
           MAX(XTKEM_MAX(JI,JJ,JK),PTKET(JI,JJ,JK))
          XPABSM_MAX(JI,JJ,JK) = MAX(XPABSM_MAX(JI,JJ,JK),PPABST(JI,JJ,JK))
        !
        IF (ZAUX_WIFF(JI,JJ,JK).GE.XWIFF_MAX(JI,JJ,JK)) THEN
           XWIFF_MAX(JI,JJ,JK) = ZAUX_WIFF(JI,JJ,JK)
           XWIDD_MAX(JI,JJ,JK) = ZAUX_WIDD(JI,JJ,JK)
        ENDIF
        !
      ENDDO
    ENDDO
  ENDDO
!
IF(LMINMAX_VORT) THEN
! Computation of vorticities (same as in the diag) 
  ZVOX(:,:,:)=GY_W_VW(PWT,XDYY,XDZZ,XDZY)-GZ_V_VW(PVT,XDZZ)
  ZVOX(:,:,2)=ZVOX(:,:,3)
  ZVOY(:,:,:)=GZ_U_UW(PUT,XDZZ)-GX_W_UW(PWT,XDXX,XDZZ,XDZX)
  ZVOY(:,:,2)=ZVOY(:,:,3)
  ZVOZ(:,:,:)=GX_V_UV(PVT,XDXX,XDZZ,XDZX)-GY_U_UV(PUT,XDYY,XDZZ,XDZY)
  ZVOZ(:,:,2)=ZVOZ(:,:,3)
  ZVOZ(:,:,1)=ZVOZ(:,:,3)
  ZVOX=MYF(MZF(MXM(ZVOX)))
  ZVOY=MZF(MXF(MYM(ZVOY)))
  ZVOZ=MXF(MYF(MZM(ZVOZ)))
  DO JK=IKB,IKE
   DO JJ=IJB,IJE
    DO JI=IIB,IIE
      XUM1_MAX(JI,JJ,JK) = MAX(XUM1_MAX(JI,JJ,JK),ZVOX(JI,JJ,JK)) ! 
      XVM1_MAX(JI,JJ,JK) = MAX(XVM1_MAX(JI,JJ,JK),ZVOY(JI,JJ,JK)) ! 
      XWM1_MAX(JI,JJ,JK) = MAX(XWM1_MAX(JI,JJ,JK),ZVOZ(JI,JJ,JK)) ! 
      XUM1_MIN(JI,JJ,JK) = MIN(XUM1_MIN(JI,JJ,JK),ZVOX(JI,JJ,JK)) ! 
      XVM1_MIN(JI,JJ,JK) = MIN(XVM1_MIN(JI,JJ,JK),ZVOY(JI,JJ,JK)) ! 
      XWM1_MIN(JI,JJ,JK) = MIN(XWM1_MIN(JI,JJ,JK),ZVOZ(JI,JJ,JK)) ! 
    END DO
   END DO
  END DO
END IF
!
! Computation of P at sea level (same as in the diag) 
!
IF(LMINMAX_MSLP) THEN
  IF(NRR > 0) THEN
!   compute the ratio : 1 + total water mass / dry air mass
    ZRV_OV_RD = XRV / XRD
    ZTHV(:,:,:) = 1. + XRT(:,:,:,1)
    DO JLOOP = 2,1+NRRL+NRRI                
      ZTHV(:,:,:) = ZTHV(:,:,:) + XRT(:,:,:,JLOOP)
    END DO
! compute the virtual potential temperature when water is present in any form
    ZTHV(:,:,:) = PTHT(:,:,:) * (1.+XRT(:,:,:,1)*ZRV_OV_RD) / ZTHV(:,:,:)
  ELSE
! compute the virtual potential temperature when water is absent
    ZTHV(:,:,:) = PTHT(:,:,:)
  END IF
!
  ZGAMREF=-6.5E-3
!  Exner function at the first mass point
  ZWORK4(:,:) = (PPABST(:,:,IKB) /XP00)**(XRD/XCPD)
!  virtual temperature at the first mass point
  ZWORK4(:,:) = ZWORK4(:,:) * ZTHV(:,:,IKB)
!  virtual temperature at ground level
  ZWORK4(:,:) = ZWORK4(:,:) - ZGAMREF*((XZZ(:,:,IKB)+XZZ(:,:,IKB+1))/2.-XZS(:,:))
!  virtual temperature at sea level
  ZMSLP(:,:) = ZWORK4(:,:) - ZGAMREF*XZS(:,:)
!  average underground virtual temperature
  ZMSLP(:,:) = 0.5*(ZWORK4(:,:)+ZMSLP(:,:))
!  surface pressure
  ZWORK4(:,:) = ( PPABST(:,:,IKB) + PPABST(:,:,IKB-1) )*.5
!  sea level pressure (hPa)
  ZMSLP(:,:) = 1.E-2*ZWORK4(:,:)*EXP(XG*XZS(:,:)/(XRD*ZMSLP(:,:)))
!
  DO JJ=IJB,IJE
    DO JI=IIB,IIE
      XMSLP_MAX(JI,JJ) = MAX(XMSLP_MAX(JI,JJ),ZMSLP(JI,JJ)) ! 
      XMSLP_MIN(JI,JJ) = MIN(XMSLP_MIN(JI,JJ),ZMSLP(JI,JJ)) ! 
    END DO
  END DO
END IF
!
!
!Computation of 10 m height wind gusts and 10 m + 20 m height TKE
IF(LMINMAX_WINDFFTKE) THEN
  JK=IKB 
  DO WHILE((JK/=IKE-1))
   IF ((XZHATM(JK+1) .GE. 10.) .AND. (XZHATM(JK)<10.)) THEN
    DO JJ=IJB,IJE
     DO JI=IIB,IIE
      ZCURRENT_ZON10M(JI,JJ)=PUT(JI,JJ,JK) + &
            (PUT(JI,JJ,JK+1)-PUT(JI,JJ,JK)) * &
            (10.-XZHATM(JK))/(XZHATM(JK+1)-XZHATM(JK))
      ZCURRENT_MER10M(JI,JJ)=PVT(JI,JJ,JK) + &
            (PVT(JI,JJ,JK+1)-PVT(JI,JJ,JK)) * &
            (10.-XZHATM(JK))/(XZHATM(JK+1)-XZHATM(JK))
      IF (CTURB/='NONE') THEN
        ZCURRENT_TKE10M(JI,JJ)=PTKET(JI,JJ,JK) + &
            (PTKET(JI,JJ,JK+1)-PTKET(JI,JJ,JK)) * &
            (10.-XZHATM(JK))/(XZHATM(JK+1)-XZHATM(JK))
      END IF
     END DO
    END DO
   END IF
   IF (CTURB/='NONE') THEN
    IF ((XZHATM(JK+1) .GE. 20.) .AND. (XZHATM(JK)<20.)) THEN
     DO JJ=IJB,IJE
      DO JI=IIB,IIE
       ZCURRENT_TKE20M(JI,JJ)=PTKET(JI,JJ,JK) + &
            (PTKET(JI,JJ,JK+1)-PTKET(JI,JJ,JK)) * &
            (20.-XZHATM(JK))/(XZHATM(JK+1)-XZHATM(JK))
      END DO
     END DO
    END IF
   END IF
   JK=JK+1
  END DO
END IF
!
! Computation of the Updraft Helicity (UH) 
IF (LUH_MAX) THEN
  JK=IKB 
  ZUH = 0.
  DO WHILE((XZHAT(JK+1) .LE. 5000))
    IF (XZHAT(JK) .GE. 2000) THEN
      DO JJ=IJB,IJE
        DO JI=IIB,IIE
          ZUH(JI,JJ)=ZUH(JI,JJ)+0.5*(XZHAT(JK+1)-XZHAT(JK))* &
                     (PWT(JI,JJ,JK+1)*ZVOZ(JI,JJ,JK+1)+PWT(JI,JJ,JK)*ZVOZ(JI,JJ,JK))
        END DO
      END DO
    END IF
   JK=JK+1
  END DO
  DO JJ=IJB,IJE
    DO JI=IIB,IIE
      XUH_MAX(JI,JJ)= MAX(XUH_MAX(JI,JJ),ZUH(JI,JJ))
    END DO
  END DO
END IF
!
IF(LMINMAX_WINDFFTKE) THEN
  DO JJ=IJB,IJE
    DO JI=IIB,IIE
      ZWORK1(JI,JJ) = SQRT(ZCURRENT_ZON10M(JI,JJ)**2+ZCURRENT_MER10M(JI,JJ)**2)
      XWMOD10MAX_MAX(JI,JJ) = MAX(XWMOD10MAX_MAX(JI,JJ),ZWORK1(JI,JJ))
      IF (CTURB/='NONE') THEN
        ZWORK2(JI,JJ) =  4. * SQRT(PTKET(JI,JJ,IKB))
        ZWORK3(JI,JJ) = 4. * SQRT(ZCURRENT_TKE10M(JI,JJ))
        ZWORK4(JI,JJ) = 3.8 * SQRT(ZCURRENT_TKE20M(JI,JJ))
        ZWORK5(JI,JJ) = ZWORK1(JI,JJ) + ZWORK2(JI,JJ)
        ZWORK6(JI,JJ) = ZWORK1(JI,JJ) + ZWORK3(JI,JJ)
        ZWORK7(JI,JJ) = ZWORK1(JI,JJ) + ZWORK4(JI,JJ)
        XTKEMAX_MAX(JI,JJ) = MAX(XTKEMAX_MAX(JI,JJ),ZWORK2(JI,JJ))
        XTKE10MAX_MAX(JI,JJ) = MAX(XTKE10MAX_MAX(JI,JJ),ZWORK3(JI,JJ))
        XTKE20MAX_MAX(JI,JJ) = MAX(XTKE20MAX_MAX(JI,JJ),ZWORK4(JI,JJ))
        XFF10MAX_MAX(JI,JJ) = MAX(XFF10MAX_MAX(JI,JJ),ZWORK5(JI,JJ))
        XFF10MAX2_MAX(JI,JJ) = MAX(XFF10MAX2_MAX(JI,JJ),ZWORK6(JI,JJ))
        XFF10MAX_AROME_MAX(JI,JJ) = MAX(XFF10MAX_AROME_MAX(JI,JJ),ZWORK7(JI,JJ))
      END IF
    END DO
  END DO
END IF
END SUBROUTINE MEAN_FIELD
